/*
 * This file is part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <errno.h>
#include <string.h>

#include "common.h"
#include "faultd-config.h"
#include "log.h"
#include "module.h"

static struct list_head modules[FAULTD_MODULE_TYPE_MAX];
static int category_disabled[FAULTD_MODULE_TYPE_MAX];

static void init_heads(void)
{
	unsigned i;

	for (i = 0; i < ARRAY_SIZE(modules); ++i) {
		INIT_LIST_HEAD(modules + i);
		category_disabled[i] = 0;
	}
}

int faultd_module_register(struct faultd_module *module)
{
	if (!module->name || !module->init) {
		log_error("Missing module name or init()");
		return -EINVAL;
	}

	if (!modules[0].next)
		init_heads();

	list_add_tail(&module->node, modules + module->type);

	return 0;
}

void faultd_module_unregister(struct faultd_module *module)
{
	list_del(&module->node);
}

int faultd_modules_init(sd_event *event_loop, struct faultd_config *conf)
{
	struct faultd_module *module, *module_to_clean;
	int i;
	int ret = 0;
	struct faultd_config *module_conf = NULL;

	if (!modules[0].next)
		init_heads();

	for (i = 0; i < (int)(ARRAY_SIZE(modules)); ++i) {
		if (category_disabled[i]) {
			log_debug("Skipping modules category %d", i);
			continue;
		}

		log_debug("Initializing modules category %d", i);
		list_for_each_entry(module, modules + i, node) {
			if (module->disabled)
				continue;
			log_debug("Initializing %s module", module->name);

			ret = faultd_config_get_subconf(conf, module->name, &module_conf);
			if (ret < 0 && ret != -ENOENT) {
				log_error_errno(ret, "Could not get subconfig for %s", module->name);
				goto err_cleanup;
			}

			ret = module->init(module, module_conf, event_loop);
			if (ret < 0) {
				log_error_errno(ret, "Could not initialize module %s", module->name);
				goto err_cleanup;
			}

			faultd_config_cleanup(module_conf);
			free(module_conf);
			module_conf = NULL;
		}
	}

	return 0;

err_cleanup:
	faultd_config_cleanup(module_conf);
	free(module_conf);

	for (; i >= 0; --i) {
		if (category_disabled[i])
			continue;

		log_debug("Cleaning up modules category %d", i);
		/* From first element of the list... */
		for (module_to_clean = list_first_entry(modules + i,
												struct faultd_module,
												node);
			 /* iterate as long as modules has been initialized */
			 module_to_clean != module &&
				 &module_to_clean->node != (modules + i);
			 /* move to next one */
			 module_to_clean = list_next_entry(module_to_clean, node)) {
			if (module_to_clean->disabled)
				continue;

			log_debug("Cleaning up %s module",
					  module_to_clean->name);
			module_to_clean->cleanup(module_to_clean);
		}
	}

	return ret;
}

void faultd_modules_cleanup()
{
	struct faultd_module *module;
	int i;

	for (i = FAULTD_MODULE_TYPE_MAX - 1; i >= 0; --i) {
		if (category_disabled[i])
			continue;

		log_debug("Cleaning up modules category %d", i);
		list_for_each_entry(module, modules + i, node) {
			if (module->disabled)
				continue;

			log_debug("Cleaning up %s module", module->name);
			module->cleanup(module);
		}
	}
}

struct faultd_module *faultd_get_module_by_name(const char *name)
{
	struct faultd_module *module;
	int i;

	for (i = FAULTD_MODULE_TYPE_MAX - 1; i >= 0; --i) {
		list_for_each_entry(module, modules + i, node) {
			if (strcmp(module->name, name) == 0)
				return module;
		}
	}

	return NULL;
}

void faultd_disable_module_category(int category)
{
	category_disabled[category] = 1;
}
