/*
 * This file is part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <errno.h>
#include <getopt.h>
#include <json-c/json.h>
#include <stdio.h>
#include <poll.h>
#include <systemd/sd-bus.h>
#include <systemd/sd-event.h>
#include <signal.h>
#include <syslog.h>
#include <unistd.h>

#include "log.h"
#include "module.h"
#include "systemd_dbus.h"
#include "faultd-config.h"

static int terminate = 0;
static struct faultd_config config;

int sigint_handler(sd_event_source *s, const struct signalfd_siginfo *si,
				   void *userdata);

int sigint_handler(sd_event_source *s,
                   const struct signalfd_siginfo *si,
                   void *userdata)
{
	sd_event *loop = userdata;

	sd_event_exit(loop, 0);
	terminate = 1;
	return 0;
}

static void print_help(const char *name)
{
	printf("Usage: %s [OPTIONS]\n"
			"Options: \n"
			"  --log-level <level>            Set log level to <level>\n"
			"  -D                             Set log level to debug\n"
			"  -d, --disable-module <module>  Disable <module>\n"
			"  --no-action                    Do not execute any action\n"
			"  -c, --config-file <file>       Read faultd configuration from <file>\n"
			"  -h, --help                     Display this help and exit\n", name);
}

static int parse_argv(int ac, char *av[])
{
	int c, r;
	enum {
		ARG_LOG_LEVEL = 0x100,
		ARG_NO_ACTION,
	};
	static const struct option options[] = {
		{"log-level", required_argument, NULL, ARG_LOG_LEVEL},
		{"disable-module", required_argument, NULL, 'd'},
		{"no-action", no_argument, NULL, ARG_NO_ACTION},
		{"config-file", required_argument, NULL, 'c'},
		{"help", no_argument, NULL, 'h'},
		{}
	};

	struct faultd_module *module;
	const char *config_file = FAULTD_DEFAULT_CONFIG_FILE;
	bool using_default_config = 1;

	while ((c = getopt_long(ac, av, "Dc:d:h", options, NULL)) >= 0) {
		switch (c) {

		case ARG_LOG_LEVEL:
			r = log_parse_level_name(optarg);
			if (r < 0) {
				fprintf(stderr, "Unknown log level: \"%s\"\n", optarg);
				return r;
			}
			log_set_max_level(r);
			break;

		case 'D':
			log_set_max_level(LOG_DEBUG);
			break;

		case 'c':
			config_file = optarg;
			using_default_config = 0;

			break;

		case 'h':
			print_help(av[0]);
			exit(0);
		case 'd':
			module = faultd_get_module_by_name(optarg);
			if (module) {
				module->disabled = 1;
				log_debug("Disabled %s module", optarg);
			} else {
				log_warning("Could not find %s module", optarg);
			}

			break;

		case ARG_NO_ACTION:
			faultd_disable_module_category(FAULTD_MODULE_TYPE_ACTION);
			break;

		default:
			print_help(av[0]);
			return -EINVAL;
		}
	}

	r = access(config_file, F_OK | R_OK);
	if (r) {
		if (!using_default_config || errno != ENOENT) {
			log_error("Could not access config file %s", config_file);
			return -errno;
		}

		log_debug("No config file");
		config.root = json_object_new_object();
		if (!config.root)
			return -ENOMEM;
	} else {
		log_debug("Using config file: %s", config_file);
		r = faultd_parse_config(config_file, &config);
		if (r < 0)
			return r;
	}

	return 0;
}

int main(int ac, char *av[])
{
	int rc;
	sd_bus *bus = NULL;
	sd_event *loop;
	sigset_t ss;

	rc = parse_argv(ac, av);
	if (rc < 0) {
		fprintf(stderr, "Failed to parse command line options: %s\n", strerror(-rc));
		return -1;
	}

	rc = faultd_acquire_systemd_bus(&bus);
	if (rc < 0) {
		log_error_errno(rc, "Failed to acquire the default system bus connection: %m");
		return -1;
	}

	rc = sd_event_new(&loop);
	if (rc < 0) {
		log_error_errno(rc, "Failed to allocate the event loop.");
		return -1;
	}

	/* If there is way to register a catch-all handler */
	/* sd_bus_attach_event(bus, loop, SD_EVENT_PRIORITY_NORMAL); */
	sigemptyset(&ss);
	sigaddset(&ss, SIGINT);
	rc = sigprocmask(SIG_BLOCK, &ss, NULL);
	rc = sd_event_add_signal(loop, NULL, SIGINT, sigint_handler, loop);


	rc = faultd_modules_init(loop, &config);
	if (rc < 0) {
		log_error("Failed to initialize modules %d.", rc);
		return -1;
	}

	printf("Hello world!\n");

	sd_event_loop(loop);

	faultd_modules_cleanup();
	faultd_config_cleanup(&config);

	sd_bus_close(bus);
	return 0;
}
