/*
 * This file is a part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "log.h"
#include "module.h"
#include "event.h"
#include "event_processor.h"
#include "faultd_started_event.h"
#include "system_booted_event.h"
#include "database.h"

static const char LAST_BOOT_ID_OID[33] = "32a4ca4f96ddfcd8a1f50269014ff320";

struct startup_listener {
	struct faultd_module module;
};

static int retrieve_last_boot_id(sd_id128_t *boot_id)
{
	faultd_oid_t oid;
	int ret;
	struct faultd_object *root;

	ret = faultd_oid_from_string(&oid, LAST_BOOT_ID_OID);
	if (ret < 0) {
		log_error("Could not build oid from string %s", LAST_BOOT_ID_OID);
		return ret;
	}

	faultd_object_new(&root);
	ret = database_get_by_oid(&oid, root);
	if (ret < 0) {
		log_debug("No boot event found");
		*boot_id = SD_ID128_NULL;
		ret = 0;
		goto out;
	}

	ret = faultd_object_get_uuid(root, "boot_id", boot_id);
	if (ret < 0) {
		log_error("could not get boot_id object");
		goto out;
	}

out:
	faultd_object_unref(root);
	return ret;
}

static void store_boot_id(sd_id128_t *boot_id)
{
	faultd_oid_t oid;
	struct faultd_object *obj;
	int ret;

	faultd_oid_from_string(&oid, LAST_BOOT_ID_OID);

	faultd_object_new(&obj);
	faultd_object_append_oid(obj, JDBIDKEYNAME, &oid);
	faultd_object_append_uuid(obj, "boot_id", boot_id);

	ret = database_store(obj, &oid);
	if (ret < 0)
		log_error_errno(ret, "Could not store object in database");
	faultd_object_unref(obj);
}

static int startup_listener_init(struct faultd_module *module,
		struct faultd_config *cfg, sd_event *event)
{
	int rc;
	struct faultd_event *started_ev, *booted_ev;
	struct sb_event_data sb_ev_data;
	sd_id128_t boot_id;

	rc = faultd_event_create(FAULTD_STARTED_EVENT_ID, NULL, &started_ev);
	if (rc) {
		log_error_errno(rc, "Unable to allocate an event");
		return rc;
	}

	rc = event_processor_report_event(started_ev);
	if (rc) {
		log_error_errno(rc, "Unable to report event");
		faultd_event_unref(started_ev);
		return rc;
	}

	rc = retrieve_last_boot_id(&boot_id);
	if (rc < 0) {
		log_error_errno(rc, "Could not retrieve boot id");
		return rc;
	}

	if (sd_id128_equal(started_ev->boot_id, boot_id))
		return 0;

	store_boot_id(&started_ev->boot_id);
	sb_ev_data.prev_boot_id = boot_id;

	rc = faultd_event_create(SYSTEM_BOOTED_EVENT_ID, &sb_ev_data, &booted_ev);
	if (rc) {
		log_error_errno(rc, "Unable to allocate an event");
		return rc;
	}

	rc = event_processor_report_event(booted_ev);
	if (rc) {
		log_error_errno(rc, "Unable to report event");
		faultd_event_unref(booted_ev);
		return rc;
	}

	return 0;
}

static void startup_listener_cleanup(struct faultd_module *module)
{
}

struct startup_listener startup_listener = {
	.module = {
		.name = "startup_listener",
		.type = FAULTD_MODULE_TYPE_LISTENER,

		.init = startup_listener_init,
		.cleanup = startup_listener_cleanup,
		.node = LIST_HEAD_INIT(startup_listener.module.node),
	},
};

FAULTD_MODULE_REGISTER(&startup_listener.module)
