/*
 * This file is part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <systemd/sd-bus.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>

#include "common.h"
#include "systemd_dbus.h"
#include "log.h"

static sd_bus *systemd_bus = NULL;

int faultd_dbus_call(char *service, char *obj, char *interface, char *method,
					 sd_bus_error *error, sd_bus_message **out_msg,
					 char *types, ...)
{
	_cleanup_(sd_bus_unrefp) sd_bus *bus = NULL;
	_cleanup_(sd_bus_message_unrefp) sd_bus_message *m = NULL;
	_cleanup_(sd_bus_error_free) sd_bus_error error_buf = SD_BUS_ERROR_NULL;
	int ret;

	if (!error)
		error = &error_buf;

	ret = faultd_acquire_systemd_bus(&bus);
	if (ret < 0)
		return ret;

	ret = sd_bus_message_new_method_call(bus, &m, service,
										 obj, interface, method);
	if (ret < 0)
		goto fail;

	if (types) {
		va_list args;

		va_start(args, types);
		ret = sd_bus_message_appendv(m, types, args);
		va_end(args);
		if (ret < 0)
			goto fail;
	}

	return sd_bus_call(bus, m, 0, error, out_msg);

fail:
	return sd_bus_error_set_errno(error, ret);
}

static int get_bus_by_address(const char *addr, sd_bus **bus)
{
	_cleanup_(sd_bus_unrefp) sd_bus *rbus = NULL;
	int ret;

	ret = sd_bus_new(&rbus);
	if (ret < 0)
		return ret;

	ret = sd_bus_set_address(rbus, addr);
	if (ret < 0)
		return ret;

	ret = sd_bus_start(rbus);
	if (ret < 0)
		return ret;

	*bus = rbus;
	rbus = NULL;
	return 0;
}

int faultd_acquire_systemd_bus(sd_bus **bus)
{
	int ret;

	if (systemd_bus) {
		*bus = systemd_bus;
		sd_bus_ref(systemd_bus);
		return 0;
	}

	ret = get_bus_by_address(SYSTEMD_KERNEL_BUS, bus);
	if (ret < 0) {
		ret = get_bus_by_address(SYSTEMD_PRIVATE_BUS, bus);
		if (ret < 0)
			return ret;
	}

	systemd_bus = *bus;

	return 0;
}
